#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "export.h"
#include "idl_tools.h"

/* function protos */
extern IDL_VPTR IDL_CDECL twain_hastwain(int argc, IDL_VPTR argv[], char *argk);
extern IDL_VPTR IDL_CDECL twain_select(int argc, IDL_VPTR argv[], char *argk);
extern IDL_VPTR IDL_CDECL twain_acquire(int argc, IDL_VPTR argv[], char *argk);

/* define the TWAIN functions */
static IDL_SYSFUN_DEF twain_functions[] = {

    { twain_hastwain,	"TWAIN_PRESENT",	0, 0, 0},
    { twain_select,	"TWAIN_SELECT",		0, 0, 0},
    { twain_acquire,	"TWAIN_ACQUIRE",	2, 5, 0},

};

/*
 *	has_twain = TWAIN_PRESENT()
 *	error = TWAIN_SELECT()
 *	error = TWAIN_ACQUIRE(info,image[,r,g,b])
 *
 *	Note: <0 = error   0 = ok
 */

#ifdef WIN32

#include <windows.h>
#include "twain.h"
#include "table.h"
#include "idl_twain_glue.h"

/* startup call when DLM is loaded */
int twain_startup(void)
{
	if (!IDL_AddSystemRoutine(twain_functions, TRUE, 
		ARRLEN(twain_functions))) {
		return IDL_FALSE;
	}
	IDL_ExitRegister(twain_exit_handler);

	
	return(IDL_TRUE);
}

/* globals */
HWND		hwnd = NULL;
TW_IDENTITY	AppIdentity;
TW_INT16	AcqFlag = 0;


/* called when IDL is shutdown */
void twain_exit_handler(void)
{
	CloseConnection();
	if (hwnd) DestroyWindow(hwnd);
}


/* local protos */
static int Setup(void);
static HGLOBAL MessageLoop(void);


static int Setup()
{
	if (hwnd) return(1);

	hwnd = CreateWindow("STATIC",                     // class
			    "Dummy",                      // title
			    WS_POPUPWINDOW,               // style
			    CW_USEDEFAULT, CW_USEDEFAULT, // x, y
			    CW_USEDEFAULT, CW_USEDEFAULT, // width, height
			    HWND_DESKTOP,                 // parent window
			    NULL,                         // hmenu
			    NULL,		          // hinst
			    NULL);                        // lpvparam

	if (!hwnd) return(0);

	AppIdentity.Id = 0; 		
	AppIdentity.Version.MajorNum = 1;
	AppIdentity.Version.MinorNum = 000;
	AppIdentity.Version.Language = TWLG_USA;
	AppIdentity.Version.Country  = TWCY_USA;
	lstrcpy (AppIdentity.Version.Info,  "IDL_TWAIN 1.0.0.0  04/18/1999");
	lstrcpy (AppIdentity.ProductName,   "IDL_TWAIN");

	AppIdentity.ProtocolMajor = 1;
	AppIdentity.ProtocolMinor = 7;
	AppIdentity.SupportedGroups =  DG_IMAGE | DG_CONTROL;
	lstrcpy (AppIdentity.Manufacturer,  "Anonymous");
	lstrcpy (AppIdentity.ProductFamily, "IDL_TOOL DLM");

	// pass app particulars to glue code
	TWInitialize (&AppIdentity, hwnd);

	return(1);
}

static HGLOBAL MessageLoop()
{
	MSG		msg;
	TW_UINT16	finished = 0;
	HGLOBAL		hDIB = NULL;

	while (GetMessage((LPMSG)&msg, NULL, 0, 0))
	{
		if (!ProcessTWMessage ((LPMSG)&msg, hwnd, &finished, &hDIB)) {
			TranslateMessage ((LPMSG)&msg);
			DispatchMessage ((LPMSG)&msg);
		} else {
			if (finished || hDIB) break;
		}
	}                     

	return(hDIB);
}

IDL_VPTR IDL_CDECL twain_hastwain(int argc, IDL_VPTR argv[], char *argk)
{
	if (!Setup()) IDL_GettmpLong(-1);

	if (TWOpenDSM() == TRUE) {
		TWCloseDSM ();
		return(IDL_GettmpLong(1));
	}

	return(IDL_GettmpLong(0));
}

IDL_VPTR IDL_CDECL twain_select(int argc, IDL_VPTR argv[], char *argk)
{
	BOOL	err;

	if (!Setup()) IDL_GettmpLong(-1);
	
	if (TWOpenDSM() == TRUE) {
		err = TWSelectDS();
		TWCloseDSM ();
		if (err) return(IDL_GettmpLong(-1));

		return(IDL_GettmpLong(0));
	}

	return(IDL_GettmpLong(-1));
}

/* Returned Info structure */
typedef struct {
	IDL_LONG	iChannels;
	IDL_LONG	iDims[2];
	float		fDPI[2];
	short		iHavePalette;
	IDL_LONG	iIndex;
	IDL_LONG	iNum;
	short		iPixelType;
	IDL_STRING	sType;
} twain_info;
static IDL_LONG	dim_dims[] = {1, 2};
static IDL_STRUCT_TAG_DEF s_tags[] = {
	{"CHANNELS", 0, (void *) IDL_TYP_LONG},
	{"DIMENSIONS", dim_dims, (void *) IDL_TYP_LONG},
	{"DPI", dim_dims, (void *) IDL_TYP_FLOAT},
	{"HAS_PALETTE", 0, (void *) IDL_TYP_INT},
	{"IMAGE_INDEX", 0, (void *) IDL_TYP_LONG},
	{"NUM_IMAGES", 0, (void *) IDL_TYP_LONG},
	{"PIXEL_TYPE", 0, (void *) IDL_TYP_INT},
	{"TYPE", 0, (void *) IDL_TYP_STRING},
	{ 0 }
};

IDL_VPTR IDL_CDECL twain_acquire(int argc, IDL_VPTR argv[], char *argk)
{
	TW_BOOL		ShowUI = TRUE;
	TW_INT16	result = TRUE;
	IDL_LONG	eRet	= -1;
	HGLOBAL		hDIB;
	IDL_LONG	i,j;

	if (!Setup()) IDL_GettmpLong(-1);

	for(i=0;i<argc;i++) IDL_EXCLUDE_EXPR(argv[i]);
                    
	/* use the default data source */
	TWDefaultDS();

	hDIB = NULL;
	AcqFlag = 0;    /* Not ready to transfer yet */
	if (TWOpenDSM () == TRUE) {
		if (TWOpenDS() == TRUE) {
			if (TWXferMech (hwnd) == TWRC_SUCCESS) {
				if (!TWIsDSEnabled()) {
						result = TWEnableDS (ShowUI);

						AcqFlag = 3;
						/*AcqFlag = 0 Do Not Accept MSG_XFERREADY
						 *          1 Disable/CloseDS/CloseDSM
						 *          2 Disable Only
						 *          3 Do Not Disable - only if ShowUI=TRUE
						 */
						if (result) hDIB = MessageLoop();
				}
			} else {
				if (TWCloseDS()) {
					if (TWCloseDSM()) {
						AcqFlag = 0;                        
					}
				}
				eRet = -2;
				goto err_out;
			}
		} else {
			eRet = -3;
			LogMessage("OpenDS failed -- TWAcquire\r\n");
			TWCloseDSM();
			goto err_out;
		}

		/*
		* Cannot Enable Source
		*/
		if (result == FALSE)  {    
			if (TWCloseDS()) {
				if (TWCloseDSM()) {
					AcqFlag = 0;                        
				}
			}
			eRet = -4;
			goto err_out;
		} 
	} else {
		eRet = -5;
		goto err_out;
	}

err_out:
	/* make extra sure things are closed down properly */
	CloseConnection();

	/* if we have no output, return an error */
	if (!hDIB) return(IDL_GettmpLong(eRet));
{
	LPBITMAPINFOHEADER	lpdib;
	RGBQUAD			*pColor;
	UCHAR			*pData,*pOut,*pSrc,*pDst;
	IDL_StructDefPtr	sdef;
	IDL_LONG		iDims[4];
	IDL_VPTR		vpTmp;
	twain_info		*info;
	IDL_LONG		iBits,iColors,iScanline;

	/* build the struct */
	sdef = IDL_MakeStruct(0, s_tags);
	iDims[0] = 1;
	info = (twain_info *)IDL_MakeTempStruct(sdef, 1, iDims, &vpTmp, TRUE);

	/* cook up return values */
	lpdib = (LPBITMAPINFOHEADER)GlobalLock(hDIB);
	if (!lpdib) {
		GlobalFree(hDIB);
		return(IDL_GettmpLong(-1));
	}
	pColor = (RGBQUAD *)((LPSTR)lpdib + (WORD)(lpdib->biSize));

	if (lpdib->biCompression != BI_RGB) {
		GlobalUnlock(hDIB);
		GlobalFree(hDIB);
		IDL_MessageFromBlock(msg_block, M_TOOLS_BADDIBFORMAT, 
				IDL_MSG_LONGJMP,"Compressed");
	} 
	if ((lpdib->biBitCount == 4) || (lpdib->biBitCount == 16)) {
		GlobalUnlock(hDIB);
		GlobalFree(hDIB);
		IDL_MessageFromBlock(msg_block, M_TOOLS_BADDIBFORMAT, 
				IDL_MSG_LONGJMP,"4bit and 16 bit");
	}

	/* raw data pointers */
	iBits = lpdib->biBitCount;
	iColors = lpdib->biClrUsed;
	if ((iColors == 0) && (iBits == 8)) iColors = 256;
	if ((iColors == 0) && (iBits == 1)) iColors = 2;
	iScanline = (lpdib->biWidth*iBits)/8;
	while(iScanline & 0x3) iScanline++;
	pData = (UCHAR *)(pColor+iColors);

	/* Info Struct */
	info->iDims[0] = lpdib->biWidth;
	info->iDims[1] = IDL_ABS(lpdib->biHeight);
	info->fDPI[0] = (float)lpdib->biXPelsPerMeter;
	info->fDPI[1] = (float)lpdib->biYPelsPerMeter;
	info->fDPI[0] *= (0.0254f);
	info->fDPI[1] *= (0.0254f);
	if (iColors) info->iHavePalette = 1;
	info->iIndex = 0;
	info->iNum = 1;
	info->iPixelType = IDL_TYP_BYTE;
	IDL_StrStore(&(info->sType),"TWAIN");
	info->iChannels = 1;
	if (iBits > 8) info->iChannels = 3;
	IDL_VarCopy(vpTmp,argv[1]);

	/* palette */
	if (info->iHavePalette) {
		if (argc >= 3) {
			pOut = (UCHAR *)IDL_MakeTempVector(IDL_TYP_BYTE,iColors,
				IDL_BARR_INI_NOP,&vpTmp);
			for(i=0;i<iColors;i++) pOut[i] = pColor[i].rgbRed;
			IDL_VarCopy(vpTmp,argv[2]);
		}
		if (argc >= 4) {
			pOut = (UCHAR *)IDL_MakeTempVector(IDL_TYP_BYTE,iColors,
				IDL_BARR_INI_NOP,&vpTmp);
			for(i=0;i<iColors;i++) pOut[i] = pColor[i].rgbGreen;
			IDL_VarCopy(vpTmp,argv[3]);
		}
		if (argc >= 5) {
			pOut = (UCHAR *)IDL_MakeTempVector(IDL_TYP_BYTE,iColors,
				IDL_BARR_INI_NOP,&vpTmp);
			for(i=0;i<iColors;i++) pOut[i] = pColor[i].rgbBlue;
			IDL_VarCopy(vpTmp,argv[4]);
		}
	}

	/* destination image */
	if (info->iChannels == 1) {
		iDims[0] = info->iDims[0];
		iDims[1] = info->iDims[1];
		iDims[4] = 2;
	} else {
		iDims[0] = info->iChannels;
		iDims[1] = info->iDims[0];
		iDims[2] = info->iDims[1];
		iDims[4] = 3;
	}
	pOut = (UCHAR *)IDL_MakeTempArray(IDL_TYP_BYTE,iDims[4],iDims, 
			IDL_BARR_INI_NOP,&vpTmp);
	IDL_VarCopy(vpTmp,argv[0]);

	switch(iBits) {
		case 1 : 
			pDst = pOut;
			for(j=0;j<info->iDims[1];j++) {
				pSrc = pData + j*iScanline;
				for(i=0;i<info->iDims[0];i++) {
					IDL_LONG	by,bi;

					by = i / 8;
					bi = 7 - (i % 8);
					*pDst++ = (pSrc[by] & (1<<bi)) ? 1 : 0;
				}
			}
			break;
		case 8:
			pDst = pOut;
			for(j=0;j<info->iDims[1];j++) {
				pSrc = pData + j*iScanline;
				for(i=0;i<info->iDims[0];i++) {
					*pDst++ = *pSrc++;
				}
			}
			break;
		case 24:
			pDst = pOut;
			for(j=0;j<info->iDims[1];j++) {
				pSrc = pData + j*iScanline;
				for(i=0;i<info->iDims[0];i++) {
					pDst[2] = *pSrc++;
					pDst[1] = *pSrc++;
					pDst[0] = *pSrc++;
					pDst += 3;
				}
			}
			break;
		case 32:
			pDst = pOut;
			for(j=0;j<info->iDims[1];j++) {
				pSrc = pData + j*iScanline;
				for(i=0;i<info->iDims[0];i++) {
					pDst[2] = *pSrc++;
					pDst[1] = *pSrc++;
					pDst[0] = *pSrc++;
					pSrc++;
					pDst += 3;
				}
			}
			break;
	}
}
	
	/* We're out of here... */
	GlobalUnlock(hDIB);
	GlobalFree(hDIB);

	return(IDL_GettmpLong(0)); 
}


#else

/* startup call when DLM is loaded */
int twain_startup(void)
{
	if (!IDL_AddSystemRoutine(twain_functions, TRUE, 
		ARRLEN(twain_functions))) {
		return IDL_FALSE;
	}
	IDL_ExitRegister(twain_exit_handler);
	return(IDL_TRUE);
}

/* called when IDL is shutdown */
void twain_exit_handler(void)
{
}

IDL_VPTR IDL_CDECL twain_hastwain(int argc, IDL_VPTR argv[], char *argk)
{
	return(IDL_GettmpLong(0));
}

IDL_VPTR IDL_CDECL twain_select(int argc, IDL_VPTR argv[], char *argk)
{
	return(IDL_GettmpLong(-1));
}

IDL_VPTR IDL_CDECL twain_acquire(int argc, IDL_VPTR argv[], char *argk)
{
	return(IDL_GettmpLong(-1));
}

#endif

